//--------------------------------------------------------------------------//
// xmtnimage70.cc                                                           //
// Latest revision: 08-22-2022                                              //
// See xmtnimage.h for Copyright Notice                                     //
//--------------------------------------------------------------------------//

#include "xmtnimage.h"

extern Globals     g;
extern Image      *z;
extern int         ci;
extern int in_pattern_recognition;
enum{ EDM, BUF };
enum{ NOTHING, SEGMENTATION, NEURAL, DIFFERENCING };
const int MAXGRAINS = 20000;          //// Change this number to enlarge
const int MAXSPOTS = 100;             //// Change this number to enlarge

extern int grain_count;               //// No. of grains found
extern int *grain;                    //// size of up to MAXGRAINS grains
extern double *grain_signal;          //// signal of each grain
extern int *grain_size;               //// size distribution
extern int *show_grain_size;          //// size distribution scaled for graph
extern int *grain_signal_distribution;//// signal distribution in 256 bins

extern int **grain_coordinate;        //// 0,1 = center x,y of each grain
                               //// 2,3 = upper left x,y of each grain bounding box
                               //// 4,5 = lower right x,y of each grain bounding box

int auditmode = 1;
int pixelthickness = 1;
int range = 1;
int want_labels = 1;
int binfac = 1;
int want_save = 0;
int want_save_dens = 0;
int want_save_output_histogram = 0;
int band_detection = 2;
int width = 85;
int height = 20;
double scale    = 1.0;
double thresh   = 0.95;
char filename[FILENAMELENGTH];
char dfilename[FILENAMELENGTH];
char ofilename[FILENAMELENGTH];
int simage = 1;
int cimage = 2;


void bin_histograms(double ***histogram, int binfac);


//--------------------------------------------------------------------------//
// audit - test if image could have been manipulated                        //
// The latest fad in science is to claim without evidence that someone has  //
// falsified a blot image. These functions will provide documentation one   //
// way or the other.                                                        //
//--------------------------------------------------------------------------//
void audit(void)
{
  static Dialog *dialog;
  static int simage = 1;
  static int cimage = 2;
  int jj, kk;

  strcpy(filename, "Histograms.dat");
  strcpy(dfilename, "Density.dat");
  strcpy(ofilename, "Output.dat");
  simage = ci;
  dialog = new Dialog;
  if(dialog==NULL){ message(g.nomemory); return; }
  strcpy(dialog->title,"Image forensics / Detect image manipulation");
  strcpy(dialog->radio[0][0],"Image manipulation test"); 
  strcpy(dialog->radio[0][1],"Histogram analysis"); 
  strcpy(dialog->radio[0][2],"Histogram analysis w/scaling"); 
  strcpy(dialog->radio[0][3],"Find differences"); 
  strcpy(dialog->radio[0][4],"Butterfly plot"); 
  strcpy(dialog->radio[0][5],"Horiz.Profile Diff.Plot"); 
  strcpy(dialog->radio[0][6],"Vert.Profile Diff.Plot"); 
  strcpy(dialog->radio[0][7],"Zebra");             
  strcpy(dialog->radio[0][8],"Smoothness map");             
  dialog->radiono[0] = 9;
  dialog->radioset[0] = auditmode;

  strcpy(dialog->radio[1][0],"Band detection"); 
  strcpy(dialog->radio[1][1],"Automatic"); 
  strcpy(dialog->radio[1][2],"Automatic - fixed size"); 
  strcpy(dialog->radio[1][3],"Manual"); 
  dialog->radiono[1] = 4;
  dialog->radioset[1] = band_detection;

  strcpy(dialog->radio[2][0],"Appearance"); 
  strcpy(dialog->radio[2][1],"Normal"); 
  strcpy(dialog->radio[2][2],"Heavy pixels"); 
  dialog->radiono[2] = 3;
  dialog->radioset[2] = pixelthickness;

  strcpy(dialog->radio[3][0],"Labels"); 
  strcpy(dialog->radio[3][1],"Boxes and labels"); 
  strcpy(dialog->radio[3][2],"Labels only"); 
  strcpy(dialog->radio[3][3],"None"); 
  dialog->radiono[3] = 4;
  dialog->radioset[3] = want_labels;

  strcpy(dialog->boxes[0],"Data");
  strcpy(dialog->boxes[1],"Suspect Image");
  strcpy(dialog->boxes[2],"Control Image (if needed)");
  strcpy(dialog->boxes[3],"Scale");
  strcpy(dialog->boxes[4],"Range (if applicable)");
  strcpy(dialog->boxes[5],"Signif. r^2 threshold");
  strcpy(dialog->boxes[6],"Pattern threshold");
  strcpy(dialog->boxes[7],"Pattern match wt");
  strcpy(dialog->boxes[8],"Pattern mismatch wt");
  strcpy(dialog->boxes[9],"Filename for histograms");
  strcpy(dialog->boxes[10],"Filename for density trace");
  strcpy(dialog->boxes[11],"Filename for output histograms");
  strcpy(dialog->boxes[12],"Spot size (for fixed-size)");
  strcpy(dialog->boxes[13],"Width");
  strcpy(dialog->boxes[14],"Height");
  strcpy(dialog->boxes[15],"Bin factor");

  dialog->boxtype[0]=LABEL;
  dialog->boxtype[1]=INTCLICKBOX;
  dialog->boxtype[2]=INTCLICKBOX;
  dialog->boxtype[3]=STRING;
  dialog->boxtype[4]=INTCLICKBOX;
  dialog->boxtype[5]=STRING;
  dialog->boxtype[6]=STRING;
  dialog->boxtype[7]=STRING;
  dialog->boxtype[8]=STRING;
  dialog->boxtype[9]=TOGGLESTRING;
  dialog->boxtype[10]=TOGGLESTRING;
  dialog->boxtype[11]=TOGGLESTRING;
  dialog->boxtype[12]=LABEL;
  dialog->boxtype[13]=STRING;
  dialog->boxtype[14]=STRING;
  dialog->boxtype[15]=INTCLICKBOX;

  dialog->boxmin[1]=1; dialog->boxmax[1]=g.image_count-1;
  dialog->boxmin[2]=1; dialog->boxmax[2]=g.image_count-1;
  dialog->boxmin[4]=1; dialog->boxmax[4]=65536;
  dialog->boxmin[15]=1; dialog->boxmax[15]=64;

  
  for(jj=0; jj<10; jj++) for(kk=0; kk<20; kk++) dialog->radiotype[jj][kk]=RADIO;
  simage = max(0, min(simage, g.image_count-1));
  cimage = max(0, min(cimage, g.image_count-1));

  itoa(simage, dialog->answer[1][0], 10);
  itoa(cimage, dialog->answer[2][0], 10);
  sprintf(dialog->answer[3][0], "%g", scale);
  sprintf(dialog->answer[4][0], "%d", range);
  sprintf(dialog->answer[5][0], "%g", thresh);
  sprintf(dialog->answer[6][0], "%g", g.pattern_thresh);
  sprintf(dialog->answer[7][0], "%g", g.pattern_pweight);
  sprintf(dialog->answer[8][0], "%g", g.pattern_bweight);

  sprintf(dialog->answer[9][0], "%d", want_save);
  strcpy(dialog->answer[9][1], filename);
  dialog->boxset[9] = want_save;

  sprintf(dialog->answer[10][0], "%d", want_save_dens);
  strcpy(dialog->answer[10][1], dfilename);
  dialog->boxset[10] = want_save_dens;

  sprintf(dialog->answer[11][0], "%d", want_save_output_histogram);
  strcpy(dialog->answer[11][1], ofilename);
  dialog->boxset[11] = want_save_output_histogram;

  //// 12 is label, skip
  sprintf(dialog->answer[13][0], "%d", width);
  sprintf(dialog->answer[14][0], "%d", height);
  sprintf(dialog->answer[15][0], "%d", binfac);

  dialog->noofradios=4;
  dialog->noofboxes=16;
  dialog->helptopic=0;  
  dialog->want_changecicb = 0;
  dialog->f1 = auditcheck;
  dialog->f2 = null;
  dialog->f3 = null;
  dialog->f4 = auditfinish;
  dialog->f5 = null;
  dialog->f6 = null;
  dialog->f7 = null;
  dialog->f8 = null;
  dialog->f9 = null;
  dialog->width = 560; 
  dialog->height = 500;
  dialog->transient = 1;
  dialog->wantraise = 0;
  dialog->radiousexy = 0;
  dialog->boxusexy = 0;
  dialog->radioextrawidth = 60;
  strcpy(dialog->path,".");
  dialog->message[0]=0;      
  dialog->busy = 0;
 
  dialogbox(dialog);

  auditmode      = dialog->radioset[0];
  band_detection = dialog->radioset[1];
  pixelthickness = dialog->radioset[2];
  want_labels    = dialog->radioset[3];
  simage = atoi(dialog->answer[1][0]);
  cimage = atoi(dialog->answer[2][0]);
  scale  = atof(dialog->answer[3][0]);
  range  = atoi(dialog->answer[4][0]);
  thresh = atof(dialog->answer[5][0]);
  g.pattern_thresh  = atof(dialog->answer[6][0]);
  g.pattern_pweight = atof(dialog->answer[7][0]);
  g.pattern_bweight = atof(dialog->answer[8][0]);
  want_save = dialog->boxset[9];
  want_save_dens = dialog->boxset[10];
  width  = atoi(dialog->answer[12][0]);
  height = atoi(dialog->answer[13][0]);
  binfac = atoi(dialog->answer[15][0]);
}


//--------------------------------------------------------------------------//
//  auditfinish                                                             //
//--------------------------------------------------------------------------//
void auditfinish(dialoginfo *a)
{   
   a=a;
}


//--------------------------------------------------------------------------//
//  auditcheck                                                              //
//--------------------------------------------------------------------------//
void auditcheck(dialoginfo *a, int radio, int box, int boxbutton)
{
  radio=radio; box=box; boxbutton=boxbutton;
  if(radio != -2) return;   //// User clicked Ok or Enter
  int simage=1, cimage=1;
  double scale;
  double thresh, pattern_threshold, pattern_pweight, pattern_bweight;

  auditmode = a->radioset[0];
  band_detection = a->radioset[1];
  pixelthickness = a->radioset[2];
  want_labels    = a->radioset[3];

  simage = atoi(a->answer[1][0]);
  cimage = atoi(a->answer[2][0]);
  scale  = atof(a->answer[3][0]);
  range  = atoi(a->answer[4][0]);
  thresh = atof(a->answer[5][0]);
  pattern_threshold = atof(a->answer[6][0]);
  pattern_pweight   = atof(a->answer[7][0]);
  pattern_bweight   = atof(a->answer[8][0]);

  want_save = a->boxset[9];
  strcpy(filename, a->answer[9][0]);

  want_save_dens = a->boxset[10];
  strcpy(dfilename, a->answer[10][0]);

  want_save_output_histogram = a->boxset[11];
  strcpy(ofilename, a->answer[11][0]);

  //// 12 is a label, skip
  width  = atoi(a->answer[13][0]);
  height = atoi(a->answer[14][0]);
  binfac = atoi(a->answer[15][0]);

  if(!between(simage, 0, g.image_count-1) ||
     !between(cimage, 0, g.image_count-1))
  {   message("Bad image number");
      return;
  }  
  if((z[simage].is_zoomed) ||
     (z[cimage].is_zoomed))
  {   message("Cannot analyze zoomed images");
      return;
  }  
  strcpy(filename, a->answer[9][1]);
  strcpy(dfilename, a->answer[10][1]);
  strcpy(ofilename, a->answer[11][1]);

  switch(auditmode)
  {   
      case 1: audit_western_histogram(simage, cimage, scale, width, height, thresh, 
              pattern_threshold, pattern_pweight, pattern_bweight, 
              band_detection, want_labels, want_save, filename, 
              want_save_output_histogram, ofilename, 0); break;
      case 2: audit_western_histogram(simage, cimage, scale, width, height, thresh, 
              pattern_threshold, pattern_pweight, pattern_bweight, 
              band_detection, want_labels, want_save, filename, 
              want_save_output_histogram, ofilename, 1); break;
      case 3: audit_find_differences(simage, cimage, scale); break;
      case 4: audit_butterfly_plot(simage, cimage, scale); break;
      case 5: audit_horiz_plot(simage, cimage, scale, dfilename); break;
      case 6: audit_vert_plot(simage, cimage, scale, dfilename); break;
      case 7: audit_zebra(simage, range); break;
      case 8: audit_smoothness_map(simage, scale, range); break;
      default: break;
  }

}



//--------------------------------------------------------------------------//
//  audit_western_histogram                                                 //
//  correlates histograms but no correction for contrast change             //
//  contrast_correction = 0 - no correction for contrast and pixel offset   //
//  contrast_correction = 1 - estimate contrast and pixel offset first      //
///-------------------------------------------------------------------------//
void audit_western_histogram(int simage, int cimage, double scale, 
     int width, int height, double thresh,
     double pattern_thresh, double pattern_pweight, double pattern_bweight,
     int band_detection, int want_labels, int want_save, char *filename, 
     int want_save_output_histogram, char *ofilename,
     int contrast_correction)
{

              //// 100 bands  
              //// 0=pixval, 1=count 
              //// 65536 possible pixvalues
    static double histogram[100][2][65536];  
    static double total[65536];           //// total pixel values in band
    static double temp[2][65536]; 
    static double newhistogram[65536]; 
    static int    sorted1[65536];         //// sorting index of histogram 1
    static int    sorted2[65536];         //// sorting index of histogram 2
//    static double h1[20];                //// 20 top histogram values (0=highest)
//    static double h2[20];                //// 20 top histogram values (0=highest)
    double contrast;                     //// contrast relative to 1st one
    double brightness;                   //// pixel offset relative to 1st one
    double q[20];                        //// lin reg params
    double ff, r2;                       //// f value and r-squared
    double mm, bb;                       //// y = mx + b
    double tvalue, prob;
    char tempstring[1024];
    char tempstring2[1024];
    int shell = 0;
    int border = 0;
    uchar *add = 0;
    int pix=0, count=0, countmin=0, countmax=0, hitcount=0, badspots=0;
    int bpp, ct, spot, jj, kk, x1=0, y1=0, x2=0, y2=0, ino=0, xsize, ysize;
    int spot1, spot2, wlabel=0, wbox=0, xx, yy, kk2, satcount = 0;
    int first1=0, first2=0, length, maxval=65536;
    double vv;
    FILE *fp;
    Widget www, scrollbar;

    //// Shut up the compiler
    cimage = cimage;
    simage = simage;
    countmax = countmax; countmin = countmin; 
    scale = scale;
    shell = shell; border = border; 
    add = add; 

    ino = ci;   
    bpp = z[ino].bpp;
    ct  = z[ino].colortype;
    xsize = z[ino].xsize;
    ysize = z[ino].ysize;
    allocate_grain_arrays();
    q[0] = 0;
    q[1] = 0;
    q[2] = 0;

    for(kk=0; kk<65536; kk++)
    {
       newhistogram[kk] = 0;
       sorted1[kk] = 0;
       sorted2[kk] = 0;
       total[kk] = 0;
    }


    //// Parameters are
    //// ino, method, g.pattern_thresh, pweight, bweight, patternsize
    //// wantlabels, wantrectangles, labelcolor, graph, densgraph

    //// Find the bands, assume western blot
    //// for want_labels: 1=both 2=labels only 3=none
    in_pattern_recognition = 1;

    g.img_only = 1;
    if(want_labels == 1){ wlabel=1; wbox=1; }
    if(want_labels == 2){ wlabel=1; wbox=0; }

    //// Put our own labels on it
    find_spots(ci, SEGMENTATION, pattern_thresh, pattern_pweight, pattern_bweight, 10, 0, 0, 0, 0, 0);
    if (grain_count > MAXSPOTS) 
    {   message("Too many bands\nEdit MAXSPOTS to change",ERROR);
        return;
    }


    for(spot=0; spot<grain_count; spot++)
    {   
         x1 = grain_coordinate[2][spot]-2;
         y1 = grain_coordinate[3][spot]-2;
         x2 = grain_coordinate[4][spot]+2;
         y2 = grain_coordinate[5][spot]+2;
         xx = (x1 + x2) / 2;
         yy = (y1 + y2) / 2;
         if(band_detection == 2)  //// automatic, fixed size
         {
              x1 = grain_coordinate[2][spot] = xx - width/2;
              y1 = grain_coordinate[3][spot] = yy - height/2;
              x2 = grain_coordinate[4][spot] = xx + width/2;
              y2 = grain_coordinate[5][spot] = yy + height/2;         
         }
         if(wlabel == 1)
         {   line(xx, yy, xx+12, yy+12, 255, 0, 0, SET, z[ci].win, ci);
             print_grain_label(spot+1, xx, yy, 0, 0, ci);
         }
         if(wbox == 1)
             box(x1, y1, x2, y2, 0, SET, z[ci].win, ci);
    }
    g.img_only = 0;

    in_pattern_recognition = 0;
    strcpy(tempstring,"Too many bands\nPlease change parameters and repeat");
    if(grain_count >= 200){ message(tempstring); return;}
    if(bpp != 16)
        sprintf(tempstring,"Convert image from %d to 16 bits/pixel for better results\n", bpp);
    if(ct != GRAY)
        strcpy(tempstring2,"Convert image from color to grayscale for better results");
    kk = 2*(bpp !=16) + (z[ino].colortype != GRAY);
    switch(kk)
    {    case 1: message(tempstring2, ERROR); break;              //// gray 
         case 2: message(tempstring, ERROR); break;               //// color
         case 3: strcat (tempstring, tempstring2); message(tempstring, ERROR); break;  //// gray + color
    }

    //// Initialize histogram array
    for(spot=0; spot<grain_count; spot++)
    {    for(kk=0; kk<65536; kk++)
         {    histogram[spot][0][kk]=kk;  
              histogram[spot][1][kk]=0; 
         }
    }

    //// Create up to 200 histograms
    //// first index of grain_coordinate
    //// 2,3 = upper left x,y of each grain bounding box
    //// 4,5 = lower right x,y of each grain bounding box
    //// Pix values go from 0 to 6554 to make less sparse
    //// Thus two pix values can be off by up to +-5 and still count as same
    
    for(spot=0; spot<grain_count; spot++)
    {
        total[spot] = 0.0;
        if(g.getout) { message("Aborted"); g.getout = 0; break; }
        for(yy=grain_coordinate[3][spot]-2; yy<=grain_coordinate[5][spot]+2; yy++)
        {   for(xx=grain_coordinate[2][spot]-2; xx<=grain_coordinate[4][spot]+2; xx++)
            {   
                 if(outside(ino, xx, yy)) continue;
                 add = z[ino].image[0][yy] + xx*g.off[bpp];    
                 pix = pixelat(add, bpp);
                 total[spot] += (double)pix;
              
                 //// convert the pixel to monochrome then to grayscale
                 if(ct != GRAY) pix = graypixel(pix, bpp); 
                 pix = convertpixel(pix, bpp, 16, 0);

                 ////  drop pixel values of in darkest portion
                 if(between(pix, 0, 65536)) 
                 {  
                      histogram[spot][1][pix] += 1.0;  
                 }
            }
         }
    }

    //// bin histograms to make linear regression easier
    //// necessary if bands are very small

    if(binfac > 1)
    {
        for(spot = 0; spot<grain_count; spot++)
        {    
             if(g.getout) { message("Aborted"); g.getout = 0; break; }
             for(kk=0; kk<65536/binfac; kk++)
             {
                  vv = 0;
                  for(jj=0; jj<binfac; jj++) 
                  {
                      vv += histogram[spot][1][kk*binfac + jj];
if(spot==0)printf("kk %d jj %d binfac %d %d %g \n   ",kk,jj,binfac, kk*binfac + jj,vv);
                  }
                  histogram[spot][1][kk] = vv;
              }
        }
        maxval /= binfac;
    }
    
    if(want_save)
    {   if ((fp=fopen(filename,"wt")) == NULL)
        {     message("Cannot create file", ERROR);
              return; 
        }
        fprintf(fp, "## Histograms from bands in file %s\n", z[ci].name);       
        fprintf(fp, "## Pixel value for scaled count\n");       
        fprintf(fp, "## Spot no: ");
        for(kk=0; kk<grain_count; kk++)
             fprintf(fp, "%d ", kk+1);       
        fprintf(fp, "%c",'\n');
        for(kk=0; kk<maxval; kk++)
        {    fprintf(fp, "%d ",kk);
             for(spot=0; spot<grain_count; spot++)
                 fprintf(fp, "%g ", histogram[spot][1][kk]);
             fprintf(fp, "%c",'\n');
        }
        fclose(fp);
    }  

    //// set up a list
    static listinfo *ll = 0;
    static char **llinfo;
    static char *lltitle;
    ll = new listinfo;
    llinfo = new char*[600];
    lltitle = new char[512];
    strcpy(lltitle, "Western blot forensics results");
    for(spot=0; spot<600; spot++){ llinfo[spot] = new char[512]; llinfo[spot][0] = 0; }


    ll->title  = lltitle;
    ll->info   = llinfo;
    ll->count  = 0;
    ll->itemstoshow = 15;
    ll->firstitem   = 0;
    ll->wantsort    = 0;
    ll->wantsave    = 1;
    ll->helptopic   = 0;
    ll->allowedit   = 1;
    ll->edittitle  = new char[100]; 
    ll->selection   = &g.crud;
    ll->width        = 0;
    ll->wantfunction = 0;
    ll->autoupdate   = 0;
    ll->deletebutton = 1;
    ll->clearbutton  = 1;
    ll->exportbutton = 0;
    ll->highest_erased = 0;
    ll->x = 0;
    ll->y = 0;
    ll->f1 = null; // use f1 to perform an action when item is selected
    ll->f2 = null; // use f2 to update list 
    ll->f3 = null;
    ll->f4 = delete_list;
    ll->f5 = null;
    ll->f6 = null;
    ll->listnumber = 3;
    ll->wc = 0;

    //// find correlations of unscaled histograms

    count = 0;    //// number of entries calculated
    hitcount = 0; //// number over threshold that go in list
    ll->count += 2;
    g.getout = 0;

    progress_bar_open(www, scrollbar,(char*)"Progress");
    if(want_save_output_histogram)
    {   if ((fp=fopen(ofilename,"wt")) == NULL)
        {     message("Cannot create file", ERROR);
              return; 
        }
    }
 
    for(spot1=0; spot1<grain_count; spot1++)
    {
        if(g.getout) { message("Aborted"); g.getout = 0; break; }
        for(spot2=0; spot2<grain_count; spot2++)
        {
            progress_bar_update(scrollbar, (double)spot1*100/(double)grain_count);

            if(want_save_output_histogram) 
                 fprintf(fp, "Pixval Histogram1 Histogram2\n");     

            if(spot1==spot2) continue;

            //// Calculate the relative contrast and brightness for each
            //// pair of spots in case they changed them
            //// total = total pixel values in spot
            //// area = area of spot

            if(contrast_correction) //// check for contrast and brightness changes
            {
                //// do a linked sort of spot 1 by count
                sort(histogram[spot1][1], sorted1, maxval);
                if(g.getout) { message("Aborted"); g.getout = 0; return; }
                for(kk=0; kk<maxval; kk++)
                {
                    temp[0][kk] = histogram[spot1][0][sorted1[kk]];
                    temp[1][kk] = histogram[spot1][1][sorted1[kk]];
                }
                for(kk=0; kk<maxval; kk++)
                {   histogram[spot1][0][kk] = temp[0][kk];
                    histogram[spot1][1][kk] = temp[1][kk];
                }
                //// do a linked sort of spot 2 by count
                sort(histogram[spot2][1], sorted2, maxval);
                for(kk=0; kk<maxval; kk++)
                {
                    temp[0][kk] = histogram[spot2][0][sorted2[kk]];
                    temp[1][kk] = histogram[spot2][1][sorted2[kk]];
                }
                for(kk=0; kk<maxval; kk++)
                {   histogram[spot2][0][kk] = temp[0][kk];
                    histogram[spot2][1][kk] = temp[1][kk];
                }
                //// linked sort is finished

                //// get rid of zeroes then make them same length
                //// also get rid of anything below some value, which is only noise
                //// Do not go below 10 

                first1 = maxval; first2 = maxval;
                for(kk=0; kk<maxval; kk++)
                    if(histogram[spot1][1][kk] > 10.0){ first1 = kk; break;}
                for(kk=0; kk<maxval; kk++)
                    if(histogram[spot2][1][kk] > 10.0){ first2 = kk; break;}
                first1 = min(first1, first2);
                length = maxval - first1;       //// new histogram size
                if(first1 == 0)  badspots++; 
    //printf("b %d %d   %d %d %d  %d\n",spot1, spot2, first1,first2,length,badspots);


                //// truncate the histograms
                for(kk=0; kk<length; kk++)
                {
                    histogram[spot1][0][kk] = histogram[spot1][0][kk+first1];
                    histogram[spot1][1][kk] = histogram[spot1][1][kk+first1];
                    histogram[spot2][0][kk] = histogram[spot2][0][kk+first1];
                    histogram[spot2][1][kk] = histogram[spot2][1][kk+first1];
                }

                //// now do a linked sort of spot 1 by pixval
                sort(histogram[spot1][0], sorted1, length);
                for(kk=0; kk<length; kk++)
                {
                    temp[0][kk] = histogram[spot1][0][sorted1[kk]];
                    temp[1][kk] = histogram[spot1][1][sorted1[kk]];
                }
                for(kk=0; kk<length; kk++)
                {   histogram[spot1][0][kk] = temp[0][kk];
                    histogram[spot1][1][kk] = temp[1][kk];
                }

                //// do a linked sort of spot 2 by pixval
                sort(histogram[spot2][0], sorted2, length);
                for(kk=0; kk<length; kk++)
                {
                    temp[0][kk] = histogram[spot2][0][sorted2[kk]];
                    temp[1][kk] = histogram[spot2][1][sorted2[kk]];
                }
                for(kk=0; kk<length; kk++)
                {   histogram[spot2][0][kk] = temp[0][kk];
                    histogram[spot2][1][kk] = temp[1][kk];
                }

                //// linear regression on pixval
                if(length < 3)
                {    message("Not enough different pixel values\nImage not suitable", ERROR);
                     free_grain_counting_arrays();
                     progress_bar_close(www);
                     return;
                }
                linreg(histogram[spot1][0], histogram[spot2][0], length, 1, q, ff, r2, tvalue, prob);
                mm = q[1]; bb = q[0];
                contrast = mm;
                brightness = bb;

                //// Scale histogram 1 to match calculated contrast & brightness

                for(kk=0; kk<maxval; kk++)
                {
                     //// brightness shifts it to the left or right
                     //// contrast scales it
                     kk2 = cint(kk * contrast) + brightness;
                     if(kk2<maxval && kk2<maxval && kk2>=0 && kk>=0)
                         newhistogram[kk] = cint(histogram[spot1][0][kk2]);
                }
                linreg(histogram[spot2][0], newhistogram, maxval, 1, q, ff, r2, tvalue, prob);
                if(want_save_output_histogram) 
                {    for(kk=0; kk<maxval; kk++)
                     { fprintf(fp, "%d %g %g\n", kk,histogram[spot2][0][kk], newhistogram[kk]);
                     }
                }
            }else
            {   
                linreg(histogram[spot1][1], histogram[spot2][1], maxval, 1, q, ff, r2, tvalue, prob);
                if(want_save_output_histogram) 
                {    for(kk=0; kk<maxval; kk++)
                     { fprintf(fp, "%d %g %g\n", kk, histogram[spot1][1][kk], histogram[spot2][1][kk]);
                     }
                }
            }
            count++;

            //// Put the hits in the table
            if(r2 >= thresh)
            {   if(contrast_correction)
                {
                    sprintf(ll->info[5 + hitcount], "%3d   %3d      %3.3e   %3.5f    %3.4f    %3.3f    %3g  %3g   %d", 
                       spot1+1, spot2+1, ff, r2, mm, bb, tvalue, prob, length);
                }
                else
                {   sprintf(ll->info[5 + hitcount], "%3d   %3d      %3.3e   %3.5f    %3.4e    %3.4e  %3.4e   %3.4e", 
                       spot1+1, spot2+1, ff, r2, q[0], q[1], mm, bb);
                }
                hitcount++;
            }
        }
    }
    if(want_save_output_histogram) fclose(fp);

    if(badspots)
    {   jj = grain_count * grain_count;
        sprintf(tempstring, "Analyzed %d out of %d spot combinations", jj - badspots, jj);
        message(tempstring, WARNING);
    }
          
    //// Add the headers at top
    sprintf(ll->info[0], "%d patterns found",grain_count);   
    sprintf(ll->info[1], "%d out of %d histograms showed possible significant correlations.", 
        hitcount, count);
    strcpy(ll->info[2], "A correlation does not necessarily mean manipulation," );
    strcpy(ll->info[3], "only that these bands may need further study." );
    if(contrast_correction)
    {
         sprintf(ll->info[4], "spot 1 spot 2  F value     R-squared  contrast  brightness   t value  prob. pixvals");
    }else
    {    sprintf(ll->info[4], "spot 1 spot 2  F value     R-squared  slope         intercept  t value  prob.");
    }

    ll->count = hitcount + 5;
    //// show the results   

    progress_bar_close(www);


    satcount = highlightsaturatedpixels(ino, 1);
    if(satcount)
    {  if(satcount == 1) strcpy(tempstring, "pixel"); else strcpy(tempstring, "pixels");
       sprintf(ll->info[5 + hitcount],"%d saturated %s (%g percent of total) detected", 
              satcount, 
              tempstring,
              100.00 *(double)satcount / ((double)xsize * (double)ysize));
       hitcount++;
       ll->count++;
    }

    list(ll);
 
    free_grain_counting_arrays();
    switchto(ino);
    z[ino].touched = 1;
    redraw(ino);

}


//--------------------------------------------------------------------------//
//  outside  bounds check                                                   //
//--------------------------------------------------------------------------//
int outside(int ino, int xx, int yy)
{
    int answer=0;
    if(xx < 0 || yy < 0) answer = 1;
    if(xx >= z[ino].xsize || yy >= z[ino].ysize) answer = 1;
    return answer;
}




//--------------------------------------------------------------------------//
//  audit_find_differences                                                  //
//  highlight areas where pixels are different in two images                //
//--------------------------------------------------------------------------//
void audit_find_differences(int simage, int cimage, double scale)
{
    int bpp = z[simage].bpp;
    uchar *add;
    int pix1, pix2, value = 0;
    int xx, yy, rr, gg, bb;

    for(yy=0; yy<z[simage].ysize; yy++) 
    for(xx=0; xx<z[simage].xsize; xx++) 
    {   if(xx >= z[cimage].xsize) continue;
        if(yy >= z[cimage].ysize) continue;
        add = z[cimage].image[0][yy] + xx*g.off[bpp];    
        pix1 = pixelat(add, bpp);
        add = z[simage].image[0][yy] + xx*g.off[bpp];    
        pix2 = pixelat(add, bpp);
        value = (pix1 - pix2)/2;
        value = int(scale * (double)value);
        rr = 128 + value / 8;
        gg = 99;
        bb = 256 - rr;
        setpixelonimg(xx, yy, rr, gg, bb, simage);
    }
    redraw(simage);
}


//--------------------------------------------------------------------------//
//  audit_butterfly_plot                                                    //
//  plot pixel value differences between two images at each x and y         //
//--------------------------------------------------------------------------//
void audit_butterfly_plot(int simage, int cimage, double scale)
{
    scale = scale;
    int bpp = z[simage].bpp;
    int ct  = z[simage].colortype;
    int shell = 0;
    int border = 0;
    int xsize = 512;
    int ysize = 512;
    uchar *add;
    int pix1, pix2, value = 0;
    int xx, yy, rr, gg, bb, outimage;
    if(newimage((char*)"Comparison_plot",0,0,xsize+30,ysize+30,24,COLOR,1,
        shell,1,PERM,0,border,0)!=OK) return;
    outimage = ci;

    for(yy=0; yy<z[outimage].ysize; yy++) 
    for(xx=0; xx<z[outimage].xsize; xx++) 
        putRGBbytes(xx, yy, outimage, 24, 255, 255, 255);

    for(yy=0; yy<z[simage].ysize; yy++) 
    for(xx=0; xx<z[simage].xsize; xx++) 
    {   if(xx >= z[cimage].xsize) continue;
        if(yy >= z[cimage].ysize) continue;

        add = z[cimage].image[0][yy] + xx*g.off[bpp];    
        pix1 = pixelat(add, bpp);
        add = z[simage].image[0][yy] + xx*g.off[bpp];    
        pix2 = pixelat(add, bpp);

        //// convert the pixels to monochrome then to grayscale
        if(ct != GRAY) pix1 = graypixel(pix1, z[simage].bpp); 
        pix1 = convertpixel(pix1, z[simage].bpp, 16, 0);
        if(ct != GRAY)  pix2 = graypixel(pix2, z[cimage].bpp); 
        pix2 = convertpixel(pix2, z[cimage].bpp, 16, 0);

        value = (pix1 - pix2)/2;
        value = max(0, min(value, z[outimage].ysize)-1);
        rr = 128 + value / 8;
        gg = 99;
        bb = 256 - rr;
        
        putRGBbytes(pix1/128, (512 - pix2/128), outimage ,24, rr, gg, bb);
        if(pixelthickness==2)
        {
           putRGBbytes(pix1/128-1, (512 - pix2/128)-1, outimage ,24, rr, gg, bb);
           putRGBbytes(pix1/128-1, (512 - pix2/128)-0, outimage ,24, rr, gg, bb);
           putRGBbytes(pix1/128-1, (512 - pix2/128)+1, outimage ,24, rr, gg, bb);
           putRGBbytes(pix1/128-0, (512 - pix2/128)-1, outimage ,24, rr, gg, bb);
           putRGBbytes(pix1/128-0, (512 - pix2/128)+1, outimage ,24, rr, gg, bb);
           putRGBbytes(pix1/128+1, (512 - pix2/128)-1, outimage ,24, rr, gg, bb);
           putRGBbytes(pix1/128+1, (512 - pix2/128)-0, outimage ,24, rr, gg, bb);
           putRGBbytes(pix1/128+1, (512 - pix2/128)+1, outimage ,24, rr, gg, bb);
        }
    }
    switchto(outimage);
    z[outimage].touched = 1;
    rebuild_display(outimage);
    redraw(outimage);
}

//--------------------------------------------------------------------------//
//  audit_horiz_plot                                                        //
//  x axis density profile difference plot of 2 images                      //
//--------------------------------------------------------------------------//
void audit_horiz_plot(int simage, int cimage, double scale, char *dfilename)
{
    FILE *fp;
    scale = scale;
    int bpp = z[simage].bpp;
    int ct  = z[simage].colortype;
    int total1=0, total2=0;
    uchar *add;
    int pix1, pix2, index, row;
    double density[2][2000];
    double d[2][2000];
    double q[20];                        //// lin reg params
    double ff, r2;                       //// f value and r-squared
    double tvalue, prob;
    int xx, yy, kk, count;
    if(cimage == simage) 
    {   message("Two different images are needed\nfor comparison", ERROR);
        return;
    }
    for(xx=0; xx<z[simage].xsize; xx++) 
    { 
        total1=0; total2=0;
        for(yy=0; yy<z[simage].ysize; yy++) 
        {   if(xx >= z[cimage].xsize) continue;
            if(yy >= z[cimage].ysize) continue;
            if(xx >= z[simage].xsize) continue;
            if(yy >= z[simage].ysize) continue;

            add = z[cimage].image[0][yy] + xx*g.off[bpp];    
            pix1 = pixelat(add, bpp);
            add = z[simage].image[0][yy] + xx*g.off[bpp];    
            pix2 = pixelat(add, bpp);

            //// convert the pixels to monochrome then to grayscale
            if(ct != GRAY) pix1 = graypixel(pix1, z[simage].bpp); 
            pix1 = convertpixel(pix1, z[simage].bpp, 16, 0);
            if(ct != GRAY) pix2 = graypixel(pix2, z[cimage].bpp); 
            pix2 = convertpixel(pix2, z[cimage].bpp, 16, 0);

            total1 += pix1;
            total2 += pix2;
        }
        index = cint(2000 * (double)xx / (double)z[simage].xsize);
        density[0][index] = total1;
        density[1][index] = total2;           
    }
    
    //// Get rid of all the zeroes caused by incomplete pixel values
    //// This would screw up the statistics
    kk=0;
    for(index=0; index<2000; index++)
    { 
        if((density[0][index]>1000) && (density[1][index]>1000))
        {    d[0][kk] = density[0][index];
             d[1][kk] = density[1][index];
             kk++;
             count=kk;
        }
    }
    if(want_save_dens)
    {   if ((fp=fopen(dfilename,"wt")) == NULL)
        {     message("Cannot create file", ERROR);
              return; 
        }
        for(kk=0; kk<count; kk++)
            fprintf(fp, "%d %g %g\n", kk, d[0][kk], d[1][kk]);
        fclose(fp);
    }

    linreg(d[0], d[1], count, 1, q, ff, r2, tvalue, prob);

//printf("%g %g %g %g %g %g\n", q[0],q[1],ff,r2,tvalue,prob);

    //// set up a list
    static listinfo *ll = 0;
    static char **llinfo;
    static char *lltitle;
    ll = new listinfo;
    llinfo = new char*[200];
    lltitle = new char[512];
    strcpy(lltitle, "Horizontal density plot statistics");
    for(row=0; row<200; row++){ llinfo[row] = new char[512]; llinfo[row][0] = 0; }
    ll->title  = lltitle;
    ll->info   = llinfo;
    ll->count  = 0;
    ll->itemstoshow = 15;
    ll->firstitem   = 0;
    ll->wantsort    = 0;
    ll->wantsave    = 1;
    ll->helptopic   = 0;
    ll->allowedit   = 1;
    ll->edittitle  = new char[100]; 
    ll->selection   = &g.crud;
    ll->width        = 0;
    ll->wantfunction = 0;
    ll->autoupdate   = 0;
    ll->deletebutton = 1;
    ll->clearbutton  = 1;
    ll->exportbutton = 0;
    ll->highest_erased = 0;
    ll->x = 0;
    ll->y = 0;
    ll->f1 = null; // use f1 to perform an action when item is selected
    ll->f2 = null; // use f2 to update list 
    ll->f3 = null;
    ll->f4 = delete_list;
    ll->f5 = null;
    ll->f6 = null;
    ll->listnumber = 4; //// unique integer
    ll->wc = 0;
    ll->count = 5;

    sprintf(ll->info[0], "Image %d vs image %d",simage, cimage);   
    strcpy(ll->info[1], "Correlation of vertical line density");
    strcpy(ll->info[2], "Count F value      R-squared    intercept    slope    tvalue  pvalue");
    sprintf(ll->info[3], "%d    %3g     %3g     %3.3e    %3.5f  %3g  %3g", 
                         count, ff,         r2,        q[0],        q[1],  tvalue, prob);
    sprintf(ll->info[4], "Density traces stored in %s",dfilename);
    //// show the results   
    list(ll);

}



//--------------------------------------------------------------------------//
//  audit_vert_plot                                                         //
//  y axis density profile difference plot of 2 images                      //
//--------------------------------------------------------------------------//
void audit_vert_plot(int simage, int cimage, double scale, char *dfilename)
{  
    FILE *fp;
    int bpp = z[simage].bpp;
    int ct  = z[simage].colortype;
    int shell = 0;
    int border = 0;
    int xsize = 512;
    int ysize = 512;
    int total1=0, total2=0;
    uchar *add;
    int pix1, pix2, value = 0;
    int xx, yy, rr, gg, bb, outimage;
    if(newimage((char*)"Comparison_plot",0,0,xsize+30,ysize+30,24,COLOR,1,
        shell,1,PERM,0,border,0)!=OK) return;
    outimage = ci;

    for(yy=0; yy<z[outimage].ysize; yy++) 
    {   for(xx=0; xx<z[outimage].xsize; xx++) 
            putRGBbytes(xx, yy, outimage, 24, 255, 255, 255);
    }

    for(yy=0; yy<z[simage].ysize; yy++) 
    { 
        total1=0; total2=0;

        for(xx=0; xx<z[simage].xsize; xx++) 
        {   if(xx >= z[cimage].xsize) continue;
            if(yy >= z[cimage].ysize) continue;

            add = z[cimage].image[0][yy] + xx*g.off[bpp];    
            pix1 = pixelat(add, bpp);
            add = z[simage].image[0][yy] + xx*g.off[bpp];    
            pix2 = pixelat(add, bpp);

            //// convert the pixels to monochrome then to grayscale
            if(ct != GRAY) pix1 = graypixel(pix1, z[simage].bpp); 
            pix1 = convertpixel(pix1, z[simage].bpp, 16, 0);
            if(ct != GRAY) pix2 = graypixel(pix2, z[cimage].bpp); 
            pix2 = convertpixel(pix2, z[cimage].bpp, 16, 0);

            total1 += pix1;
            total2 += pix2;

            value = (pix1 - pix2)/2;
            value = g.maxvalue[bpp]/2 + int(scale * (double)value);
            value /=128;
            value = max(0, min(value, z[outimage].ysize)-1);
            rr = 128 + value / 8;
            gg = 99;
            bb = 256 - rr;

            putRGBbytes(xx, pix2/128, outimage ,24, rr, gg, bb);
            if(want_save_dens)
            {   if ((fp=fopen(dfilename,"wt")) == NULL)
                {     message("Cannot create file", ERROR);
                      return; 
                }
                fprintf(fp, "%d %d %d\n", yy, total1, total2);
                fclose(fp);
            }
        }
    }
    switchto(outimage);
    z[outimage].touched = 1;
    rebuild_display(outimage);
    redraw(outimage);
}



//--------------------------------------------------------------------------//
//  audit_zebra                                                             //
//  set alternate pixel values to black or white                            //
//--------------------------------------------------------------------------//
void audit_zebra(int simage, int range)
{ 
   int ii,jj,pix,even=0,rr,gg,bb;
   int xsize = z[simage].xsize;
   int ysize = z[simage].ysize;
   int bpp = z[simage].bpp;
               //// range is distance from central pixel
               //// iscale is contrast factor for detecting
   uchar *add;
   for(jj=0; jj<ysize; jj++)
   for(ii=0; ii<xsize; ii++)
   {   add = z[simage].image[0][jj] + ii*g.off[bpp];    
       pix = pixelat(add, bpp);

       //// decide even or odd
//     if(pix == (range+1) * int(pix / (range+1))) even=1; else even=0;

       //// decide half the pixel values are above 'range'
       //// if range is 1 this reduces to even or odd

       if(pix % (range*2) >= range) even=1; else even=0;
       if(even){ rr = 255; gg = 255; bb = 255; }
       else { rr = 0; gg = 0; bb = 0;}

       if(!outside(simage,ii,jj)) 
          setRGBbytes(z[simage].img[jj]+ii*g.off[g.bitsperpixel], 
          rr, gg, bb, g.bitsperpixel);
   }
   redraw(simage);
   printf("Hit Ctrl-R to restore image\n");

}


//--------------------------------------------------------------------------//
//  audit_smoothness_map                                                    //
//  highlight areas that have been digitally smoothed                       //
//--------------------------------------------------------------------------//
void audit_smoothness_map(int simage, double iscale, int range)
{
   int ii,jj,xx,yy,pix,pix2;
   double sharpness;
   int rr,gg,bb;
   int xsize = z[simage].xsize;
   int ysize = z[simage].ysize;
   int bpp = z[simage].bpp;
               //// range is distance from central pixel
               //// iscale is contrast factor for detecting
   uchar *add;
   for(jj=range; jj<ysize-range; jj++)
   for(ii=range; ii<xsize-range; ii++)
   {   sharpness = 0;
       add = z[simage].image[0][jj] + ii*g.off[bpp];    
       pix = pixelat(add, bpp);
       for(yy=jj-range; yy<jj+range; yy++)      
       for(xx=ii-range; xx<ii+range; xx++)      
       {
           add = z[simage].image[0][yy] + xx*g.off[bpp];    
           pix2 = pixelat(add, bpp);
           sharpness += (double)abs(pix - pix2);
       }
       sharpness /= (double)iscale*(double)iscale;
       sharpness = pow((double)sharpness, 1.3);
       bb = iscale*sharpness/100;
       bb = min(255, max(0, bb));
       rr = min(255, max(0, 255 - bb));
       gg = 0;
       if(!outside(simage,xx,yy)) 
           setRGBbytes(z[simage].img[jj]+ii*g.off[g.bitsperpixel], 
           rr, gg, bb, g.bitsperpixel);
   }
   redraw(simage);
printf("Hit Ctrl-R to restore image\n");
}
